/******************************************************************************

   heavy.c

   Show heavy process status for *nix (public version 1.1)

   Programed by Daisuke Nishikawa
   e-mail:daisuke@icrr.u-tokyo.ac.jp

   Note:
    This program dependent operation system.

   This is free software. This program is applied GNU General Public License.

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define  VERSION  "Show heavy process status, public version 1.1"
#define  USER_LEN 12
#define  USER_MAX 32
#define  BUFSIZE  256
#define  PS_BUF   256
#define  PSCMD    "ps axu"    /* System dependent */
#if !defined EXIT_SUCCESS
   #define EXIT_SUCCESS 0
   #define EXIT_FAILURE !0
#endif

typedef struct psValue_s
{
   char     user[USER_LEN];
   int      proc,heavy;
   double   cpu,mem;
} psValue;

typedef struct psFile_s
{
   FILE  *fp;
   char  *psBuf[PS_BUF];
   int   count;
} psFile;

int   FlgPs,FlgSum,FlgList;

psFile   *openPs(char *pscmd);
int   closePs(psFile *fp);
void  readPs(psValue *userList,double cpu_min,double mem_min,psFile *ps);
void  showPs(psValue *userList,psFile *ps);
void  addList(psValue *list,psValue *one);
void  sumList(psValue *list);
void  showUser(psValue *list);
void  showList(psValue *list,int flg);
void  usage(void);

int   main(int argc,char *argv[])
{
   psFile   *fp;
   int      count=0,i;
   double   cpu,mem;
   psValue  userList[USER_MAX];
   char     *pscmd=PSCMD;


   if (argc<2) {
      fputs("Usage:heavy cpu mem\n",stderr);
      return EXIT_FAILURE;
   }
   FlgPs=1;
   FlgSum=FlgList=0;
   cpu=mem=-1;
   i=1;
   while (i < argc) {
      if (*argv[i] == '-') {
	 switch (*(argv[i]+1)) {
	 case 'l':
	    FlgSum=FlgList=1;
	    FlgPs=0;
	    break;
	 case 's':
	    FlgSum=1;
	    FlgList=FlgPs=0;
	    break;
	 case 'v':
	    printf("%s\n",VERSION);
	    return EXIT_SUCCESS;
	 case 'h':
	    usage();
	    return EXIT_SUCCESS;
	 case 'p':
	    i++;
	    pscmd=argv[i];
	    break;
	 default:
	    break;
	 }
      } else {
	 switch (count) {
	 case 0:
	    sscanf(argv[i],"%lf",&cpu);
	    mem=100;
	    break;
	 case 1:
	    sscanf(argv[i],"%lf",&mem);
	    break;
	 default:
	    break;
	 }
	 count++;
      }
      i++;
   }
   if ((fp=openPs(pscmd)) == NULL) {
      fprintf(stderr,"%s: cannot run %s\n",argv[0],pscmd);
      return EXIT_FAILURE;
   }
   readPs(userList,cpu,mem,fp);
   showPs(userList,fp);
   closePs(fp);

   return EXIT_SUCCESS;
}


/*
 * Open ps process
 *
 * return : FILE*  NULL=Error
 */
psFile   *openPs(char *pscmd)
{
   psFile   *fp;

   fp=(psFile *)calloc(1,sizeof(psFile));
   if (fp != NULL) {
      fp->fp=popen(pscmd,"r");
      fp->count=0;
   }
   return fp;
}


/*
 * Close ps process
 *
 * in     : char *psBuf[]     Buffer for ps strings
 *        : int  count        Number of ps strings
 *        : FILE *fp
 * return : int               0=Success, EOF=EOF or Error
 */
int   closePs(psFile *fp)
{
   int   i;

   for (i=0;i<fp->count;i++) {
      free(fp->psBuf[i]);
   }
   return pclose(fp->fp);
}


/*
 * Read process status
 *
 * in : FILE   *fin              Input file
 *      double cpu_min,mem_min   Minimum value of heavy process
 */
void  readPs(psValue *userList,double cpu_min,double mem_min,psFile *ps)
{
   char     buf[BUFSIZE];
   psValue  value;
   int      count,c;
   size_t   len;

   count=ps->count;
   userList->proc=0;
   while (fgets(buf,BUFSIZE,ps->fp) != NULL) {     /* Get Header Line */
      if (strncmp(buf,"USER",4)==0)
	 break;
   }
   len=strlen(buf)+1;
   ps->psBuf[count]=(char *)malloc(len);
   strcpy(ps->psBuf[count],buf);
   count++;
   while (fgets(buf,BUFSIZE,ps->fp) != NULL) {
      value.heavy=0;
      if (sscanf(buf,"%s%*d%lf%lf",value.user,&value.cpu,&value.mem) != 3) {
	 for (;;) {
	    c=fgetc(ps->fp);
	    if ((c == '\n') || (c == EOF))
	       break;
	 }
	 continue;
      }
      if (value.cpu>cpu_min || value.mem>mem_min) {
	 value.heavy=1;
	 if (FlgPs==1) {
	    len=strlen(buf)+1;
	    ps->psBuf[count]=(char *)malloc(len+1);
	    strcpy(ps->psBuf[count],buf);
	    if (len == BUFSIZE) {
	       *(ps->psBuf[count]+len-1)='\n';
	       *(ps->psBuf[count]+len)='\0';
	    }
	    count++;
	 }
      }
      if (FlgSum==1)
	 addList(userList,&value);
   }
   ps->count=count;
}


/*
 * Show heavy process
 *
 * in : psValue   *userList
 *      psFile    *ps
 */
void  showPs(psValue *userList,psFile *ps)
{
   int   i;

   if (FlgPs==1) {
      if (ps->count > 1) {
	 for (i=0;i<ps->count;i++)
	    printf("%s",ps->psBuf[i]);
      } else {
	 printf("No heavy process.\n");
      }
   }
   if (FlgSum==1) {
      sumList(userList);
      showList(userList,FlgList);
   }
}


/*
 * Add psValue to list
 *
 * in     : psValue *one
 * in,out : psValue *list  List of psValue
 */
void  addList(psValue *list,psValue *one)
{
   while (list->proc>0) {
      if (strcmp(list->user,one->user)==0) {
	 list->proc++;
	 list->heavy += one->heavy;
	 list->cpu   += one->cpu;
	 list->mem   += one->mem;
	 return;
      }
      list++;
   }
   strcpy(list->user,one->user);
   list->proc++;
   list->heavy =  one->heavy;
   list->cpu   =  one->cpu;
   list->mem   =  one->mem;
   list++;
   list->proc  = 0;
}


/*
 * Sum of list
 *
 * in,out : psValue *list  List of psValue
 */
void  sumList(psValue *list)
{
   int      heavy,proc;
   double   cpu,mem;

   heavy=proc=0;
   cpu=mem=0.0;
   while (list->proc>0) {
      heavy+=list->heavy;
      proc +=list->proc;
      cpu  +=list->cpu;
      mem  +=list->mem;
      list++;
   }
   strcpy(list->user,"*");
   list->heavy=heavy;
   list->proc =proc;
   list->cpu  =cpu;
   list->mem  =mem;
   list++;
   list->proc =0;
}


/*
 * Show list of psValue
 *
 * in : psValue *list   List of psValue
 *    : int     flg     1=show all list, 0=show only last one
 */
void  showList(psValue *list,int flg)
{
   const char *Form[]={"User     Heavy/Proc   %%CPU  %%Mem\n",
		       "%-9s %4d %4d  %5.1f %5.1f\n"};

   printf(Form[0]);
   if (flg==1) {
      while (list->proc>0) {
	 printf(Form[1],list->user,list->heavy,list->proc,list->cpu,list->mem);
	 list++;
      }
   } else {
      while (list->proc>0)
	 list++;
      list--;
      printf(Form[1],list->user,list->heavy,list->proc,list->cpu,list->mem);
   }
}


/*
 * Show usage
 */
void  usage(void)
{
   printf("Usage:heavy cpu mem\n"
	  " -l     List\n"
	  " -s     Sum\n"
	  " -p ps  Ps command\n"
	  " -v     Version\n"
	  " -h     Help\n");
}
